#!perl -W

use warnings;
use XML::LibXML;

my $top_srcdir=shift @ARGV;

$top_srcdir=".." if(not defined $top_srcdir);

die if not $top_srcdir;
my$BUILDER="code automatically created by $0";

# build accessors for attribute
map{eval("sub att$_ {shift->getAttribute(\"$_\");}");} qw(name datatypefamily datatypename onrewrite default info);
sub evalue
{
  my$evalue;
  foreach ((shift)->findnodes("evalue")) {push @$evalue,$_->getAttribute ("name");}
  return $evalue;
}
sub eltparent
{
  my$element=shift;
  if($element->parentNode->nodeName eq "element")
  {return $element->parentNode;}
}
sub attribute_size
{
  return "list"  if((shift)->parentNode->nodeName eq "list");
  return "scalar";
}
sub isuid { return (shift)->parentNode->nodeName eq "uid"; }
sub attribute_uid_format { return (shift)->parentNode->getAttribute ("format")||""; }
sub attrOnduplicate { return (shift)->parentNode->getAttribute ("onduplicate")||""; }
sub attribute_element { return (shift)->parentNode->parentNode; }
sub attribute_parent { return &eltparent(&attribute_element(shift)); }
sub eltname { return &attribute_element(shift)->getAttribute ("name"); }
sub eltattribute { return (shift)->findnodes("(uid/attribute|scalar/attribute|list/attribute)"); }
sub parserXml
{
  my $admsParser=shift;
  my $myElementArray;
  my $myElementHash;
  my @admsStructure=$admsParser->getDocumentElement->findnodes("//element");
  foreach (@admsStructure)
  {
    my$element=$_;
    push @$myElementArray,$element;
    $myElementHash->{${element}->getAttribute("name")}=$element;
  }
  return ($myElementArray,$myElementHash);
}

my$SVN;
my$cygpath_top_srcdir=$top_srcdir;
if($cygpath_top_srcdir=`cygpath -ad $top_srcdir 2>/dev/null`)
{
  chomp $cygpath_top_srcdir;
  $cygpath_top_srcdir="\"$cygpath_top_srcdir\"";
}
$SVN=`svnversion -n $cygpath_top_srcdir 2>/dev/null` or $SVN="unknown";
print "svn version: $SVN\n";

sub buildsymbolname
{
  my($attribute,$method)=(shift,shift,shift);
  return join"_",("adms",&eltname($attribute),$method,&attname($attribute));
}
sub buildsupersymbolname
{
  my($element,$parent,$method)=(shift,shift,shift);
  return join"_",("adms",$element->getAttribute("name"),$method,$parent->getAttribute("name"));
}

# implement attdatatypefamily/attdatatypename
sub attribute_datatypenameImpl
{
  my$attribute=shift;
  my$eltname=&eltname($attribute);
  my$attributedatatypefamily=&attdatatypefamily($attribute);
  my$aname=&attname($attribute);
  my$adatatype=&attdatatypename($attribute);
  my$attrTypeImpl;
  if(${attributedatatypefamily} eq "basicpointer")
  {
    if(${adatatype} eq "quark") {$attrTypeImpl="const char*"}
    elsif(${adatatype} eq "void") {$attrTypeImpl="void*"}
    elsif(${adatatype} eq "list") {$attrTypeImpl="p_slist*"}
    elsif(${adatatype} eq "base:character:array") {$attrTypeImpl="char**"}
  }
  elsif(${attributedatatypefamily} eq "basictype")
  {
    if(${adatatype} eq "base:integer") {$attrTypeImpl="int"}
    elsif(${adatatype} eq "base:enumeration") {$attrTypeImpl="int"}
    elsif(${adatatype} eq "base:real") {$attrTypeImpl="double"}
    elsif(${adatatype} eq "base:character:array") {$attrTypeImpl="char*"}
  }
  elsif(${attributedatatypefamily} eq "reference")
  {
    if(${adatatype} eq "*") {$attrTypeImpl="p_adms"}
    else{$attrTypeImpl="p_${adatatype}"}
  }
  else
  {die;}
  return $attrTypeImpl;
}
sub containerTypeImpl
{
  my$attribute=shift;
  return (&attribute_size($attribute) eq "list")?"p_slist":&attribute_datatypenameImpl($attribute);
}

# build method templates for each attribute 
my$FUNCPREFIX="adms_ELTNAME_list_ATTRNAME";
my$LISTNAME="mymyELTNAME->_ATTRNAME";
my$templates={
prepend=>qq#
  H win32_interface void ${FUNCPREFIX}_prepend (p_ELTNAME mymyELTNAME,ATTRCONTAINER myATTRNAME);
  C void
  C ${FUNCPREFIX}_prepend (p_ELTNAME mymyELTNAME,ATTRCONTAINER myATTRNAME)
  C {
  C   adms_slist_push(&$LISTNAME,adms_adms(myATTRNAME));
  C }
  C #,
prepend_by_id=>qq#
  H win32_interface ATTRCONTAINER ${FUNCPREFIX}_prepend_by_id (p_ELTNAME mymyELTNAME,ATTRID_H);
  C ATTRCONTAINER
  C ${FUNCPREFIX}_prepend_by_id (p_ELTNAME mymyELTNAME,ATTRID_H)
  C {
  C   ATTRCONTAINER refATTRNAME=adms_ATTRVAL_new(ATTRID_C);
  C   adms_slist_push(&$LISTNAME,adms_adms(refATTRNAME));
  C   return refATTRNAME;
  C }
  C #,
prepend_by_id_once_or_ignore=>qq#
  H win32_interface ATTRCONTAINER ${FUNCPREFIX}_prepend_by_id_once_or_ignore (p_ELTNAME mymyELTNAME,ATTRID_H);
  C ATTRCONTAINER
  C ${FUNCPREFIX}_prepend_by_id_once_or_ignore (p_ELTNAME mymyELTNAME,ATTRID_H)
  C {
  C   p_slist list=$LISTNAME;
  C   ATTRCONTAINER refATTRNAME=adms_ATTRVAL_new(ATTRID_C);
  C   while(list)
  C   {
  C     if(!adms_ATTRVAL_cmp(adms_ATTRVAL(list->data),refATTRNAME))
  C     {
  C       adms_ATTRVAL_free (refATTRNAME);
  C       return adms_ATTRVAL(list->data);
  C     }
  C     list=list->next;
  C   }
  C   adms_slist_push(&$LISTNAME,adms_adms(refATTRNAME));
  C   return refATTRNAME;
  C }
  C #,
prepend_by_id_once_or_abort=>qq#
  H win32_interface ATTRCONTAINER ${FUNCPREFIX}_prepend_by_id_once_or_abort (p_ELTNAME mymyELTNAME,ATTRID_H);
  C ATTRCONTAINER
  C ${FUNCPREFIX}_prepend_by_id_once_or_abort (p_ELTNAME mymyELTNAME,ATTRID_H)
  C {
  C   p_slist list=$LISTNAME;
  C   ATTRCONTAINER refATTRNAME=adms_ATTRVAL_new(ATTRID_C);
  C   while(list)
  C   {
  C     if(!adms_ATTRVAL_cmp(adms_ATTRVAL(list->data),refATTRNAME))
  C     {
  C       adms_message_fatal(("ELTNAME=[%s] ATTRNAME=[%s] already defined\\n",
  C         adms_ELTNAME_uid(mymyELTNAME),adms_ATTRVAL_uid(refATTRNAME)))
  C       adms_ATTRVAL_free (refATTRNAME);
  C     }
  C     list=list->next;
  C   }
  C   adms_slist_push(&$LISTNAME,adms_adms(refATTRNAME));
  C   return refATTRNAME;
  C }
  C #,
prepend_once_or_ignore=>qq#
  H win32_interface void ${FUNCPREFIX}_prepend_once_or_ignore (p_ELTNAME mymyELTNAME,ATTRCONTAINER myATTRNAME);
  C void
  C ${FUNCPREFIX}_prepend_once_or_ignore (p_ELTNAME mymyELTNAME,ATTRCONTAINER myATTRNAME)
  C {
  C   p_slist list=$LISTNAME;
  C   while(list)
  C   {
  C     if(!adms_ATTRVAL_cmp(adms_ATTRVAL(list->data),myATTRNAME))
  C       return;
  C     list=list->next;
  C   }
  C   adms_slist_push(&$LISTNAME,adms_adms(myATTRNAME));
  C }
  C #,
prepend_once_or_abort=>qq#
  H win32_interface void ${FUNCPREFIX}_prepend_once_or_abort (p_ELTNAME mymyELTNAME,ATTRCONTAINER myATTRNAME);
  C void
  C ${FUNCPREFIX}_prepend_once_or_abort (p_ELTNAME mymyELTNAME,ATTRCONTAINER myATTRNAME)
  C {
  C   p_slist list=$LISTNAME;
  C   while(list)
  C   {
  C     if(!adms_ATTRVAL_cmp(adms_ATTRVAL(list->data),myATTRNAME))
  C     {
  C       adms_message_fatal(("ELTNAME=[%s] ATTRNAME=[%s] already defined\\n",
  C         adms_ELTNAME_uid(mymyELTNAME),adms_ATTRVAL_uid(myATTRNAME)))
  C       return;
  C     }
  C     list=list->next;
  C   }
  C   adms_slist_push(&$LISTNAME,adms_adms(myATTRNAME));
  C }
  C #,
};

my$admsParser=XML::LibXML->new->parse_file ("$top_srcdir/adms.xml");

my($myElementArray,$myElementHash)=&parserXml($admsParser);
foreach(@$myElementArray) {&mkNew($_);&mkNewimpl($_);&mkFree($_);}
foreach(@$myElementArray) {&mkCode($_)}
foreach("") {&mkCode($_)}

#template for file naming: Element -> elementElement
sub makeFileName
{
  my$name=shift;
  $name=~s/^(.)/element\U$1/;
  return $name;
}

#template for macro adms_..._free
sub mkFree ()
{
  my$element=$_;
  my$ename=$element->getAttribute("name");
  $FREE{$ename}{H}.="win32_interface inline void adms_${ename}_free(p_$ename my$ename);\n";
  $FREE{$ename}{C}.="inline void adms_${ename}_free(p_$ename my$ename)\n";
  $FREE{$ename}{C}.="{\n";
  $FREE{$ename}{C}.="  if(!my$ename)\n";
  $FREE{$ename}{C}.="    return;\n";
  my$parent;
  if($parent=&eltparent($element))
  {
    my$parentname=$parent->getAttribute("name");
    if($parentname eq "subexpression")
    {
      $FREE{$ename}{C}.="  adms_enumeration_free(my$ename->_subexpression._math._dependency);\n";
      $FREE{$ename}{C}.="  free(my$ename->_subexpression._math._value);\n";
    }
    elsif($parentname eq "math")
    {
      $FREE{$ename}{C}.="  adms_enumeration_free(my$ename->_math._dependency);\n";
      $FREE{$ename}{C}.="  free(my$ename->_math._value);\n";
    }
  }
  foreach(&eltattribute($element))
  {
    my$aname=&attname($_);
    my$attributedatatypefamily=&attdatatypefamily($_);
    my$attributesize=&attribute_size($_);
    my$adatatype=&attdatatypename($_);
    if(${attributesize} eq "list")
    {
      $FREE{$ename}{C}.="  adms_slist_free(my$ename->_${aname});\n";
    }
  }
  foreach(&eltattribute($element))
  {
    my$aname=&attname($_);
    my$attributedatatypefamily=&attdatatypefamily($_);
    my$attributesize=&attribute_size($_);
    my$adatatype=&attdatatypename($_);
    if(${adatatype} eq "base:character:array")
    {
      $FREE{$ename}{C}.="  free(my$ename->_${aname});\n";
    }
  }
  $FREE{$ename}{C}.="  adms_global_increment_nb${ename}destroy();\n" if $ename eq "admst";
  $FREE{$ename}{C}.="  free(my$ename);\n";
  $FREE{$ename}{C}.="}\n";
}
#template for adms_..._new_impl
sub mkNewimpl ()
{
  my$element=$_;
  my@attribute=&eltattribute($element);
  my$ename=$element->getAttribute("name");
  my$Hdec;
  my$Cdec;
  my@arguments;
  my$i=0;
  $NEWimpl{$ename}{C}.="p_adms adms_${ename}_new_impl (p_adms arguments[])\n";
  $NEWimpl{$ename}{C}.="{\n";
  $NEWimpl{$ename}{C}.="  p_${ename} mynew${ename};\n";
  foreach(@attribute)
  {
    my$aname=&attname($_);
    my$attributedatatypefamily=&attdatatypefamily($_);
    my$adatatype=&attdatatypename($_);
    my$isuid=&isuid($_);
    if($isuid)
    {
      my$adatatypeImpl=&attribute_datatypenameImpl($_);
      my$attributedatatypefamily=&attdatatypefamily($_);
      my$adatatype=&attdatatypename($_);
      if((${attributedatatypefamily} eq "basictype")&&(${adatatype} eq "base:character:array"))
      {$NEWimpl{$ename}{C}.="  const ${adatatypeImpl} my$aname=(${adatatypeImpl})arguments[$i];\n";}
      elsif((${attributedatatypefamily} eq "basictype")&&(${adatatype} eq "base:integer"))
      {$NEWimpl{$ename}{C}.="  const ${adatatypeImpl} my$aname=ADMS2INT(arguments[$i]);\n";}
      elsif((${attributedatatypefamily} eq "basictype")&&(${adatatype} eq "base:real"))
      {$NEWimpl{$ename}{C}.="  ${adatatypeImpl} my$aname=atof((char*)arguments[$i]);\n";}
      elsif((${attributedatatypefamily} eq "reference")&&(${adatatype} eq "quark"))
      {$NEWimpl{$ename}{C}.="  ${adatatypeImpl} my$aname=adms_quark_new((char*)arguments[$i]);\n";}
      else
      {$NEWimpl{$ename}{C}.="  ${adatatypeImpl} my$aname=(${adatatypeImpl})arguments[$i];\n";}
      push @arguments,"my$aname";
      $i++;
    }
  }
  $NEWimpl{$ename}{C}.="  mynew${ename}=adms_${ename}_new(".(join",",@arguments).");\n";
  $NEWimpl{$ename}{C}.="  return adms_adms(mynew${ename});\n";
  $NEWimpl{$ename}{C}.="}\n";
  $NEWimpl{$ename}{H}.="p_adms adms_${ename}_new_impl (p_adms arguments[])";
}
#template for adms_..._new
sub mkNew ()
{
  my$element=$_;
  my@attribute=&eltattribute($element);
  my$ename=$element->getAttribute("name");
  my$Hcode;
  my$Ccode;
  my$args;
  my$hasuid;
  foreach(@attribute)
  {
    my$aname=&attname($_);
    my$attributedatatypefamily=&attdatatypefamily($_);
    my$adatatype=&attdatatypename($_);
    my$isuid=&isuid($_);
    if($isuid)
    {
      my$adatatypeImpl=&attribute_datatypenameImpl($_);
      if((${attributedatatypefamily} eq "basictype")&&(${adatatype} eq "base:character:array"))
      {push @$Hcode,"const ${adatatypeImpl} my$aname";}
      else
      {push @$Hcode,"${adatatypeImpl} my$aname";}
      push @$Ccode,"my$aname";
      push @$args,"my$aname";
      $hasuid=1;
    }
  }
  if($hasuid)
  {
    $NEW{$ename}{Cdec}=(join ",",@$Ccode);
    $NEW{$ename}{Hdec}=(join ",",@$Hcode);
    $NEW{$ename}{dec}= "p_$ename adms_${ename}_new(".($NEW{${ename}}{Hdec}).")";
    $NEW{$ename}{args}=join",",@$args;
  }
  else
  {
    $NEW{$ename}{Cdec}="";
    $NEW{$ename}{Hdec}="";
    $NEW{$ename}{dec}= "p_$ename adms_${ename}_new(void)";
    $NEW{$ename}{args}="";
  }
}
sub vla {my$e=shift; return !(
($e eq "adms")
||
($e eq "lexval")
||
($e eq "yaccval")
||
($e eq "admst")
||
($e eq "transform")
||
($e eq "itransform")
||
($e eq "admstvariable")
||
($e eq "pparse")
||
($e eq "ptraverse")
||
($e eq "text")
||
($e eq "path")
);}
sub mkCode ()
{
  my$element=shift;
  my$ename=$element?$element->getAttribute("name"):"class";
  my @C;
  if($ename eq "class")
  {
    my @admsH_H;
    push @admsH_H,"\n/* ------- $BUILDER -------------- */\n";
    push @admsH_H,"\n";
    push @admsH_H,"#ifndef element${ename}_h\n";
    push @admsH_H,"#define element${ename}_h\n";
    push @admsH_H,"\n";
    push @admsH_H,"#define SVN \"${SVN }\" /* svn release version */\n";
    push @admsH_H,"\n";
    push @admsH_H,q@
/*headers -- depend on compiler, OS, ...*/
#  if defined(_MSC_VER)
#    define PACKAGE_NAME "adms"
#    if ! defined(PACKAGE_VERSION)
#      pragma message(__FILE__ ": warning: #warning undefined macro 'PACKAGE_VERSION' - defaulted to 'unknown'")
#      define PACKAGE_VERSION "unknown"
#    endif
#    define PACKAGE_STRING PACKAGE_NAME " " PACKAGE_VERSION
#    define PACKAGE_TARNAME PACKAGE_NAME
#    define PACKAGE_BUGREPORT "r29173\@users.sourceforge.net"
#    include <direct.h>
#    define isnan _isnan
#    define adms_NAN sqrt(-1.0)
#    define ADMS_PATH_SEPARATOR "\\\\"
#    define ADMS_F_OK 0
#    ifndef inline
#      define inline
#    endif
#    define HAVE_FLOAT_H 1
#    define HAVE_STDLIB_H 1 
#    define HAVE_PUTENV 1 
#    define HAVE_LOCALE 1 
#    define HAVE_STRING_H 1 
#    define HAVE_SYS_STAT_H 1
#    include <io.h>
#  else
#    include <config.h>
#    define adms_NAN (0.0/0.0)
#    define ADMS_PATH_SEPARATOR "/"
#    include <unistd.h>
#    define ADMS_F_OK F_OK
#  endif
#  include <math.h>
#  if defined(HAVE_SYS_STAT_H)
#    include <sys/stat.h>
#  endif
#  if defined(HAVE_LOCALE_H)
#    include <locale.h>
#  endif
#  if defined(HAVE_FLOAT_H)
#    include <float.h>
#  endif
#  if defined(HAVE_UNISTD_H)
#    include <unistd.h>
#  endif
#  if defined(HAVE_STDLIB_H)
#    include <stdlib.h>
#  endif
#  if defined(HAVE_STRING_H)
#    include <string.h>
#  endif
#  include <assert.h>
#  include <signal.h>
#  include <stdarg.h>
#  include <stdio.h>
#  include <ctype.h>
#  include <errno.h>
#  include <time.h>
/* in case not Posix */
#  if defined(_S_IFDIR)
#    define ADMS_S_IFDIR _S_IFDIR
#  else 
#    define ADMS_S_IFDIR S_IFDIR
#  endif

/* check OS */
#  if defined(__CYGWIN__)
#    define ADMS_OS_MS
#    define ADMS_OS_MSCYGWIN
#    define ADMS_OS "MSCYGWIN"
#  elif defined(__MSDOS__)
#    define ADMS_OS_MS
#    define ADMS_OS_MSDOS
#    define ADMS_OS "MSDOS"
#  elif defined(_WIN64)
#    define ADMS_OS_MS
#    define ADMS_OS_MSWIN64
#    define ADMS_OS "MSWIN64"
#  elif defined(_WIN32)
#    define ADMS_OS_MS
#    define ADMS_OS_MSWIN32
#    define ADMS_OS "MSWIN32"
#  else
#    define ADMS_OS_UNKNOWN
#    define ADMS_OS "UNKNOWN"
#  endif

/* check compiler */
#  if defined(__MINGW32__)
#    ifndef WIN32
#      define WIN32
#    endif
#    define ADMS_COMPILER_MINGCC
#    define ADMS_COMPILER "MINGCC"
#  elif defined(__CYGWIN__)
#    ifndef WIN32
#      define WIN32
#    endif
#    define ADMS_COMPILER_GCC 
#    define ADMS_COMPILER "GCC" 
#  elif defined(_MSC_VER)
#    ifndef WIN32
#      define WIN32
#    endif
#    define ADMS_COMPILER_MSVC
#    define ADMS_COMPILER "MSVC"
#  else
#    define ADMS_COMPILER_CC
#    define ADMS_COMPILER "CC"
#  endif

#define ktk\
  T(pparsecloseE)\
  T(pparseaddm)\
  T(pparseeos)\
  T(pparsedollar)\
  T(pparsearoba)\
  T(pparsecroix)\
  T(pparseposition)\
  T(pparseid)\
  T(pparseuid)\
  T(pparseopenE)\
  T(pparsenot)\
  T(pparseinteger)\
  T(pparseor)\
  T(pparseisdefault)\
  T(pparseaddp)\
  T(pparsedotdot)\
  T(pparsereal)\
  T(pparseadmst)\
  T(pparsereturned)\
  T(pparseequal)\
  T(pparsecount)\
  T(pparsestar)\
  T(pparsecomma)\
  T(pparsecloseCond)\
  T(pparseident)\
  T(pparsespace)\
  T(pparseopenCond)\
  T(pparseindex)\
  T(pparsenotequal)\
  T(pparseseparator)\
  T(pparsequotedstring)\
  T(pparsereverse)\
  T(pparsesort)\
  T(pparsenilled)\
  T(pparsenumber)\
  T(pparsestring)\
  T(pparseupper_case)\
  T(pparselower_case)\
  T(pparsecontains)\
  T(pparseexists)\
  T(pparsetext)\
  T(pparseand)\
  T(pparsedot)\
  T(pparsetrue)\
  T(pparsefalse)\
  T(pparseplus)\
  T(pparseminus)\
  T(1)\
  T(A)\
  T(D)\
  T(E)\
  T(G)\
  T(M)\
  T(OPdependent)\
  T(P)\
  T(T)\
  T(TemperatureDependent)\
  T(lower_case)\
  T(upper_case)\
  T(a)\
  T(absolute)\
  T(abstol)\
  T(access)\
  T(addm)\
  T(addp)\
  T(adms)\
  T(admsmain)\
  T(admst)\
  T(return)\
  T(admstvariable)\
  T(alias)\
  T(analog)\
  T(analogfunction)\
  T(and)\
  T(any)\
  T(arg1)\
  T(arg2)\
  T(arg3)\
  T(argc)\
  T(arguments)\
  T(argv)\
  T(aroba)\
  T(array)\
  T(assignment)\
  T(attribute)\
  T(base)\
  T(block)\
  T(blockvariable)\
  T(br1)\
  T(br2)\
  T(branch)\
  T(branchalias)\
  T(builtin)\
  T(bw_and)\
  T(bw_equl)\
  T(bw_equr)\
  T(bw_not)\
  T(bw_or)\
  T(bw_xor)\
  T(c)\
  T(callfunction)\
  T(case)\
  T(caseitem)\
  T(class)\
  T(closeCond)\
  T(closeE)\
  T(code)\
  T(column)\
  T(comma)\
  T(condition)\
  T(conditional)\
  T(constant)\
  T(contains)\
  T(contribution)\
  T(count)\
  T(croix)\
  T(curline)\
  T(currentdate)\
  T(d)\
  T(datatype)\
  T(datatypefamily)\
  T(datatypename)\
  T(dbg_mem)\
  T(dbg_vla)\
  T(dbg_xml)\
  T(ddtNature)\
  T(ddx)\
  T(ddxprobe)\
  T(debug)\
  T(default)\
  T(defaultcase)\
  T(definition)\
  T(dependency)\
  T(derivate)\
  T(developer)\
  T(diagonal)\
  T(digital)\
  T(direction)\
  T(discipline)\
  T(dollar)\
  T(dot)\
  T(dotdot)\
  T(dynamic)\
  T(else)\
  T(enumeration)\
  T(eos)\
  T(equ)\
  T(equal)\
  T(error)\
  T(evaluation)\
  T(exclude)\
  T(exists)\
  T(expression)\
  T(external)\
  T(f)\
  T(false)\
  T(fatal)\
  T(filename)\
  T(flicker)\
  T(flickernoise)\
  T(flow)\
  T(forblock)\
  T(forloop)\
  T(fpos)\
  T(fromstring)\
  T(fs)\
  T(fullname)\
  T(function)\
  T(global_final)\
  T(global_instance)\
  T(global_model)\
  T(global_past)\
  T(ground)\
  T(grounded)\
  T(gt)\
  T(gt_equ)\
  T(h)\
  T(hasVoltageDependentFunction)\
  T(hasspecialnumber)\
  T(hint)\
  T(id)\
  T(ident)\
  T(idtNature)\
  T(if)\
  T(include)\
  T(index)\
  T(infboundtype)\
  T(infexpr)\
  T(info)\
  T(initial)\
  T(inout)\
  T(input)\
  T(insource)\
  T(instance)\
  T(instancenode)\
  T(itransform)\
  T(itransforminsideapplytemplate)\
  T(instanceparameter)\
  T(instantiator)\
  T(integer)\
  T(internal)\
  T(is_any)\
  T(is_neg_one)\
  T(is_one)\
  T(is_zero)\
  T(isdefault)\
  T(isrelative)\
  T(isstate)\
  T(istrue)\
  T(item)\
  T(jacobian)\
  T(k)\
  T(lexval)\
  T(lhs)\
  T(linear)\
  T(list)\
  T(local)\
  T(location)\
  T(lt)\
  T(lt_equ)\
  T(m)\
  T(mapply_binary)\
  T(mapply_ternary)\
  T(mapply_unary)\
  T(math)\
  T(maxsize)\
  T(minsize)\
  T(minus)\
  T(model)\
  T(module)\
  T(multdiv)\
  T(multmod)\
  T(multtime)\
  T(n)\
  T(name)\
  T(nature)\
  T(nilled)\
  T(nnode)\
  T(no)\
  T(node)\
  T(nodealias)\
  T(nodefrominstantiator)\
  T(nodefrommodule)\
  T(nonlinear)\
  T(not)\
  T(notequ)\
  T(notequal)\
  T(number)\
  T(obsolete)\
  T(openCond)\
  T(openE)\
  T(or)\
  T(output)\
  T(p)\
  T(package_bugreport)\
  T(package_name)\
  T(package_string)\
  T(package_tarname)\
  T(package_version)\
  T(parameter)\
  T(parameterset)\
  T(parametertype)\
  T(path)\
  T(plus)\
  T(pnode)\
  T(position)\
  T(potential)\
  T(pparse)\
  T(probe)\
  T(ps)\
  T(ptraverse)\
  T(quark)\
  T(quotedstring)\
  T(range)\
  T(real)\
  T(returned)\
  T(reverse)\
  T(rhs)\
  T(row)\
  T(scalar)\
  T(scalingunit)\
  T(scope)\
  T(seen)\
  T(separator)\
  T(setinevaluate)\
  T(setinfinal)\
  T(setingetprev)\
  T(setininitial_step)\
  T(setininstance)\
  T(setinmodel)\
  T(setinnoise)\
  T(shiftl)\
  T(shiftr)\
  T(shot)\
  T(simulator)\
  T(sizetype)\
  T(sort)\
  T(source)\
  T(space)\
  T(sr1)\
  T(star)\
  T(static)\
  T(string)\
  T(subexpression)\
  T(supboundtype)\
  T(supexpr)\
  T(switch)\
  T(template)\
  T(terminal)\
  T(text)\
  T(then)\
  T(thermal)\
  T(tmp)\
  T(tmp1)\
  T(tostring)\
  T(transform)\
  T(tree)\
  T(true)\
  T(type)\
  T(u)\
  T(uid)\
  T(unique_id)\
  T(units)\
  T(update)\
  T(usage)\
  T(usedinevaluate)\
  T(usedinfinal)\
  T(usedininitial_step)\
  T(usedininstance)\
  T(usedinmodel)\
  T(usedinnoise)\
  T(value)\
  T(valueof)\
  T(variable)\
  T(vcount)\
  T(verbose)\
  T(vlast)\
  T(void)\
  T(warning)\
  T(while)\
  T(whileblock)\
  T(whileloop)\
  T(whitenoise)\
  T(yaccval)\
  T(yes)
#define T(tk) admse_##tk,
typedef enum {
  admse__any=0,
  admse__basicinteger,
  admse__constant,
  admse__p, /*%p*/
  admse__s, /*%s*/
  admse__path,
  admse__absolute,
  admse__arguments,
  admse__aroba,
  admse__croix,
  admse__datatypename,
  admse__dollarvalue,
  admse__empty,
  admse__star,
  admse__dlist,
  admse__elist,
  admse__returnedname,
  admse__returnedvalue,
  ktk
  admse__
} admse;
@;
    push @admsH_H,"#undef win32_interface\n";
    push @admsH_H,"#if defined(WIN32)\n";
    push @admsH_H,"#  if defined(_inside_admsElement)\n";
    push @admsH_H,"#    define win32_interface __declspec(dllexport)\n";
    push @admsH_H,"#  else\n";
    push @admsH_H,"#    define win32_interface __declspec(dllimport)\n";
    push @admsH_H,"#  endif\n";
    push @admsH_H,"#else\n";
    push @admsH_H,"#  define win32_interface extern\n";
    push @admsH_H,"#endif\n";
    push @admsH_H,"#define ADMS2INT(a) ((int)(long)(a))\n";
    push @admsH_H,"#define INT2ADMS(a) ((p_adms)(long)(a))\n";
    push @admsH_H,"\n";
    foreach(@$myElementArray)
    {
      my$element=$_;
      my$ename=$element->getAttribute("name");
      push @admsH_H,"typedef struct s_$ename  t_$ename;\n";
      push @admsH_H,"typedef t_$ename*        p_$ename;\n";
      push @admsH_H,"typedef const p_$ename   p_k$ename;\n";
    }
    my($codeH,$codeC)=&miscellaneous_code;
    push @admsH_H,$codeH;
    push @admsH_H,"win32_interface char* adms_admsxml_uid (p_kadms myitem);\n";
    push @admsH_H,"#endif /* element${ename}_h */\n";
    open FILE,">$top_srcdir/admsElement/admsH.h"; print FILE (join "",@admsH_H); close FILE;

    my@admsC_H;
    push @admsC_H,"\n/* ------- $BUILDER -------------- */\n";
    push @admsC_H,"\n";
    push @admsC_H,"#ifndef adms_h\n";
    push @admsC_H,"#define adms_h\n";
    push @admsC_H,"\n";
    push @admsC_H,"#include <admsElement/admsH.h>\n";
    foreach(@$myElementArray)
    {
      my$element=$_;
      my $ename=$element->getAttribute("name");
      push @admsC_H,"#include <admsElement/".makeFileName($ename).".h>\n";
    }
    push @admsC_H,"#endif /* adms_h */\n";
    open FILE,">$top_srcdir/admsElement/admsC.h"; print FILE (join "",(@admsC_H)); close FILE;
  }
  else
  {
    my @H;
    push @H,"\n/* ------- $BUILDER -------------- */\n";
    push @H,"\n";
    push @H,"#ifndef element${ename}_h\n";
    push @H,"#define element${ename}_h\n";
    push @H,"#include <admsElement/admsH.h>\n";
    my$parent=&eltparent($element);
    if($parent)
    {
      my$parentName=$parent->getAttribute ("name");
      push @H,"#include <admsElement/".makeFileName(${parentName}).".h>\n";
    }
    push @H,$hEnum if $hEnum;
    push @H,"#define adms_$ename(member) ((p_$ename) (member))\n";
    push @H,&mkElementStructure($element);
    my@attribute=&eltattribute($element);
    push @H,"\n";
    push @H,"win32_interface char*adms_${ename}_uid (p_$ename my$ename);\n";
    if($ename ne "adms")
    {
      push @H,"win32_interface ".$NEW{$ename}{dec}.";\n";
      push @H,$FREE{$ename}{H};
    }
    if(vla($ename))
    {
      push @H,"win32_interface int adms_${ename}_cmp (p_$ename my$ename,p_$ename ref$ename);\n";
      push @H,"win32_interface ".$NEWimpl{$ename}{H}.";\n";
    }
    push @H,&mkAttributeValueto($element,"h");
    push @H,"\n";
    foreach(@attribute)
    {
      my$aname=&attname($_);
      my$attributedatatypefamily=&attdatatypefamily($_);
      my$adatatype=&attdatatypename($_);
      my$isuid=&isuid($_);
      if((${attributedatatypefamily} eq "reference")&&(${adatatype} ne "*"))
      {
        push @H,"#include <admsElement/".makeFileName(${adatatype}).".h>\n" if not $Halreadyin{$ename}{${adatatype}};
        $Halreadyin{$ename}{${adatatype}}=1;
      }
    }
    push @H,"#endif /* element${ename}_h */\n";
    open FILE,">$top_srcdir/admsElement/".makeFileName($ename).".h"; print FILE (join "",(@H)); close FILE;
    push @C,"\n/* ------- $BUILDER -------------- */\n";
    push @C,"#include <admsElement/".makeFileName($ename).".h>\n";
    push @C,"#include <admsElement/".makeFileName("admsmain").".h>\n";
    push @C,"\n";
    push @C,$FREE{$ename}{C};
    push @C,&mkElementCmp($element) if vla(${ename});
    push @C,&mkElementNew($element) if($ename ne "adms");
    push @C,&mkAttributeValueto($element,"c");
    push @C,$cEnum if $cEnum;
    if($ename eq "transform")
    {
      my($codeH,$codeC)=&miscellaneous_code;
      push @C,$codeC;
      ($codeH,$codeC)=&build_admsxml_accessors;
      push @H,$codeH;
      push @C,$codeC;
    }
    open FILE,">$top_srcdir/admsElement/".makeFileName($ename).".c"; print FILE (join "",(@C)); close FILE;
  }
}
sub mkElementStructure ()
{
  my$element=$_;
  my$ename=$element->getAttribute("name");
  my$structure.="struct s_$ename {\n";
  my$parent=&eltparent($element);
  if($parent)
  {
    my$parentName=$parent->getAttribute ("name");
    $structure.="  t_$parentName _$parentName;\n";
  }
  my@attribute=&eltattribute($element);
  foreach(@attribute)
  {
    my$aname=&attname($_);
    my$containerTypeImpl=&containerTypeImpl($_);
    if(($aname eq "datatypename")||($aname eq "pseudodatatype"))
    {
      $structure.="  admse _${aname};\n";
    }
    else
    {
      $structure.="  $containerTypeImpl "."_${aname};\n";
      $structure.="  int _${aname}_isdefault : 2;\n" if vla($ename);
    }
  }
  $structure.="};\n";
  return $structure;
}
sub mkAttributeValueto_cb
{
  my$codename=shift;
  my$attribute=shift;
  my$ename=shift;
  my$attributedatatypefamily=&attdatatypefamily($attribute);
  my$aname=&attname($attribute);
  my$adatatype=&attdatatypename($attribute);
  my$adatatypeImpl=&attribute_datatypenameImpl($attribute);
  my$code=$templates->{$codename};
  die $codename if not $code;
  $code=~s/ELTNAME/$ename/g;
  $code=~s/ATTRNAME/${aname}/g;
  $code=~s/ATTRVAL/${adatatype}/g;
  $code=~s/ATTRCONTAINER/${adatatypeImpl}/g;
  $code=~s/ATTRID_H/$NEW{${adatatype}}{Hdec}/g;
  $code=~s/ATTRID_C/$NEW{${adatatype}}{Cdec}/g;
  $code=~s/^\n//;
  my$c=$code; $c=~s/\n?\s*H.*(\n|$)//g;
  my$h=$code; $h=~s/\n?\s*C.*(\n|$)//g;
  $c=~s/(\n?)\s*C /$1/g;
  $h=~s/(\n?)\s*H /$1/g;
  return("$h\n",$c);
}
sub mkAttributeValueto ()
{
  my$element=shift;
  my$type=shift;
  my$ename=$element->getAttribute("name");
  my$Hcode;
  my$Ccode;
  my@attribute=&eltattribute($element);
  foreach(@attribute)
  {
    my$aname=&attname($_);
    my$attributesize=&attribute_size($_);
    my$attributedatatypefamily=&attdatatypefamily($_);
    my$adatatype=&attdatatypename($_);
    my$adatatypeImpl=&attribute_datatypenameImpl($_);
    my$attrOnduplicate=&attrOnduplicate($_);
    my$attributeonrewrite=&attonrewrite($_);
    my$containerTypeImpl=&containerTypeImpl($_);
    my$valueto=&buildsymbolname($_,"valueto");
    my$modifier="";
    if((${attributesize} ne "list")&&(${attributedatatypefamily} eq "basictype")&&(${adatatype} eq "base:character:array"))
    {$modifier="const ";}
    $Hcode.="win32_interface inline void ${valueto} (p_$ename my$ename,$modifier${containerTypeImpl} _${aname});\n";
    $Ccode.="\ninline void\n${valueto} (p_$ename my$ename,$modifier$containerTypeImpl _${aname})\n{\n";
    if((${attributesize} ne "list")&&(${attributedatatypefamily} eq "reference"))
    {}
    if((${attributesize} ne "list")&&(${attributedatatypefamily} eq "basictype")&&(${adatatype} eq "base:character:array"))
    {
      if(vla($ename))
      {
        $Ccode.="  if(my$ename->_${aname}_isdefault!=-1)\n";
        $Ccode.="    free(my$ename->_${aname});\n";
      }
      $Ccode.="  my$ename->_${aname}=adms_kclone(_${aname});\n";
      if(vla($ename))
      {
        $Ccode.="  my$ename->_${aname}_isdefault=0;\n";
      }
    }
    else
    {
      $Ccode.="  my$ename->_${aname}=_${aname};\n";
      $Ccode.="  my$ename->_${aname}_isdefault=0;\n" if vla($ename);
    }
    $Ccode.="}\n";
    if(${attributesize} eq "list")
    {
      my($h,$c);
      if(($attrOnduplicate eq "")||(${adatatype} eq "*")||(${adatatype} eq "adms"))
      {
        ($h,$c)=&mkAttributeValueto_cb("prepend",$_,$ename);
      }
      elsif(${attributedatatypefamily} eq "reference")
      {
        my$attributeelement=$myElementHash->{${adatatype}};
        my@codeItem;
        foreach(&eltattribute($attributeelement))
        {
          if(&isuid($_))
          {
            my$cur="ref${adatatype}->_".&attname($_);
            my$ref="my".&attname($_);
            if((&attdatatypefamily($_) eq "basictype")&&(&attdatatypename($_) eq "base:character:array"))
              {push @codeItem,"!strcmp($cur,$ref)";}
            elsif((&attdatatypefamily($_) eq "reference")&&(&attdatatypename($_) eq "quark"))
              {push @codeItem,"!strcmp($cur->_value,$ref->_value)";}
            else
              {push @codeItem,"($cur==$ref)";}
          }
        }
        my$cmp=join "&&\nC        ",@codeItem;
        my$lookup_by_id=qq#
        H win32_interface ${adatatypeImpl} adms_${ename}_list_${aname}_lookup_by_id (p_${ename} mymy$ename,$NEW{${adatatype}}{Hdec});
        C ${adatatypeImpl}
        C adms_${ename}_list_${aname}_lookup_by_id (p_$ename mymy$ename,$NEW{${adatatype}}{Hdec})
        C {
        C   p_slist list=mymy$ename->_$aname;
        C   while(list) {
        C     p_${adatatype} ref${adatatype}=adms_${adatatype}(list->data);
        C     if($cmp)
        C       return ref${adatatype};
        C     list=list->next;
        C   }
        C   return NULL;
        C }
        C #;
        $lookup_by_id=~s/^\n//;
        my$C=$lookup_by_id; $C=~s/\n?\s*H.*(\n|$)//g;
        my$H=$lookup_by_id; $H=~s/\n?\s*C.*(\n|$)//g;
        $C=~s/(\n?)\s*C /$1/g;
        $H=~s/(\n?)\s*H /$1/g;
        $c.=$C; $h.=$H."\n";
        if(
          (($ename =~ "assignment|block|branch|contribution|module|node|transform|variable")&&($aname =~ "attribute"))
          ||
          (($ename =~ "module")&&($aname =~ "block"))
        )
        {
          ($H,$C)=&mkAttributeValueto_cb("prepend",$_,$ename);
          $c.=$C; $h.=$H;
        }
        if(
          (($ename =~ "adms")&&($aname =~ "attribute|variable"))
          ||
          (($ename =~ "admsmain")&&($aname =~ "variable"))
          ||
          (($ename =~ "itransform")&&($aname =~ "variable"))
          ||
          (($ename =~ "module")&&($aname =~ "assignment|blockvariable|contribution|expression"))
        )
        {
          ($H,$C)=&mkAttributeValueto_cb("prepend_by_id",$_,$ename);
          $c.=$C; $h.=$H;
        }
        if(($ename eq "module")&&($aname =~ "branch|node|probe|source"))
        {
          my($H,$C)=&mkAttributeValueto_cb("prepend_by_id_once_or_ignore",$_,$ename);
          $c.=$C; $h.=$H;
        }
        if(($ename eq "module")&&($aname =~ "branch"))
        {
          ($H,$C)=&mkAttributeValueto_cb("prepend_once_or_ignore",$_,$ename);
          $c.=$C; $h.=$H;
        }
        if(
          (($ename =~ "admsmain")&&($aname =~ "module|nature"))
          ||
          (($ename =~ "module")&&($aname =~ "branchalias|instance|node|nodealias|range"))
        )
        {
          my($H,$C)=&mkAttributeValueto_cb("prepend_by_id_once_or_abort",$_,$ename);
          $c.=$C; $h.=$H;
        }
        if(
          (($ename =~ "admsmain")&&($aname =~ "discipline"))
          ||
          (($ename =~ "analogfunction|block|module")&&($aname =~ "variable"))
          ||
          (($ename =~ "instance")&&($aname =~ "parameterset|terminal"))
          ||
          (($ename =~ "module")&&($aname =~ "analogfunction"))
          ||
          (($ename =~ "variable")&&($aname =~ "alias"))
        )
        {
          my($H,$C)=&mkAttributeValueto_cb("prepend_once_or_abort",$_,$ename);
          $c.=$C; $h.=$H;
        }
      }
      else
        {die;}
      $Hcode.=$h;
      $Ccode.=$c;
    }
  }
  return $Hcode if($type eq "h");
  return $Ccode if($type eq "c");
}
sub mkElementCmp ()
{
  my$element=$_;
  my$ename=$element->getAttribute("name");
  my$code;
  my$codeItem;
  $code.="int adms_${ename}_cmp (p_$ename my$ename,p_$ename ref$ename)\n";
  $code.="{\n";
  $code.="  int status;\n";
  my@attribute=&eltattribute($element);
  foreach(@attribute)
  {
    my$aname=&attname($_);
    my$attributesize=&attribute_size($_);
    my$attributedatatypefamily=&attdatatypefamily($_);
    my$adatatype=&attdatatypename($_);
    my$isuid=&isuid($_);
    if($isuid)
    {
      my$cur="my$ename->_".&attname($_);
      my$ref="ref$ename->_".&attname($_);
      if((${attributedatatypefamily} eq "basictype")&&(${adatatype} eq "base:character:array"))
        {push @$codeItem,"!!strcmp($cur,$ref)";}
      elsif((${attributedatatypefamily} eq "reference")&&(${adatatype} eq "quark"))
        {push @$codeItem,"!!strcmp($cur->_value,$ref->_value)";}
      else
        {push @$codeItem,"($cur != $ref)";}
    }
  }
  $code.="  status =\n";
  $code.="  (\n    ";
  if($codeItem)
  {
    $code.=join "\n    ||\n    ",@$codeItem if $codeItem;
  }
  else
  {
    $code.=1;
  }
  $code.="\n  );\n";
  $code.="  return status;\n";
  $code.="}\n";
  return $code;
}
sub mkElementNew ()
{
  my$element=$_;
  my$ename=$element->getAttribute("name");
  my$code;
  my$codeItem;
  my@uidattributes;
  my@attribute=&eltattribute($element);
  foreach(@attribute)
  {
    my$aname=&attname($_);
    my$attributesize=&attribute_size($_);
    my$adatatype=&attdatatypename($_);
    my$attributedatatypefamily=&attdatatypefamily($_);
    my$isuid=&isuid($_);
    my$adefault=&attdefault($_);
    my$valueto=&buildsymbolname($_,"valueto");
    if($isuid)
    {
      push @$codeItem,"mynew$ename->_${aname}_isdefault=-1;" if vla($ename);
      push @$codeItem,"$valueto (mynew$ename,my$aname);";
      push @uidattributes,$_;
      push @$codeItem,"mynew$ename->_${aname}_isdefault=1;" if vla($ename);
    }
    elsif(defined ${adefault})
    {
      my $defval;
      push @$codeItem,"mynew$ename->_${aname}_isdefault=-1;" if vla($ename);
      if(${adatatype} eq "base:enumeration")
      { $defval="mynew$ename->_${aname}=admse_${adefault};"; }
      elsif(${adatatype} eq "base:integer")
      { $defval="mynew$ename->_${aname}=${adefault};" }
      elsif(${adatatype} eq "base:real")
      { $defval="mynew$ename->_${aname}=${adefault};" }
      elsif(${adatatype} eq "base:character:array")
      { $defval="mynew$ename->_${aname}=adms_kclone(\"${adefault}\");"; }
      elsif(${adatatype} eq "quark")
      { $defval.="adms_${ename}_valueto_${aname}(mynew$ename,adms_quark_new(\"${adefault}\"));"; }
      elsif(${adatatype} eq "integer")
      { $defval.="adms_${ename}_valueto_${aname}(mynew$ename,adms_integer_new(${adefault}));"; }
      elsif(${adatatype} eq "real")
      { $defval.="adms_${ename}_valueto_${aname}(mynew$ename,adms_real_new(${adefault}));"; }
      elsif(${adatatype} eq "enumeration")
      {
        $defval.="adms_${ename}_valueto_${aname}(mynew$ename,adms_enumeration_new(admse_${adefault}));";
      }
      elsif(${adefault} eq "")
      { $defval.="adms_${ename}_valueto_${aname}(mynew$ename,NULL);"; }
      push @$codeItem,$defval;
      push @$codeItem,"mynew$ename->_${aname}_isdefault=1;" if vla($ename);
    }
    else
    {
      push @$codeItem,"mynew$ename->_${aname}_isdefault=-1;" if vla($ename);
    }
  }
  $code.=$NEW{$ename}{dec}."\n";
  $code.="{\n";
  $code.="  p_$ename mynew$ename=(p_$ename) calloc(1,(size_t)sizeof(t_$ename));\n";
  $code.="  adms_global_increment_nb${ename}new();\n" if($ename eq "admst");
  my$parent=&eltparent($element);
  if($parent)
  {
    $code.=&parent_new($element,"mynew$ename->",$element,$parent);
  }
  $code.="  ";
  $code.=join "\n  ",@$codeItem if $codeItem;
  $code.="\n";
  $code.="  return mynew$ename;\n";
  $code.="}\n";
  if(vla($ename))
  {
    $code.=$NEWimpl{$ename}{C};
  }
  my$count=1;
  my$format=$element->findnodes("uid")->get_node(0)->getAttribute ("format");
  my$enum;
  foreach(@uidattributes)
  {
    my$item;        
    my$aname=&attname($_);
    my$attributedatatypefamily=&attdatatypefamily($_);
    my$adatatype=&attdatatypename($_);
    if($format =~ m/\$$count/)
    {
      if(${adatatype} eq "base:character:array") {$item="mynew$ename->_$aname";}
      elsif(${adatatype} eq "base:enumeration") {die;}
      elsif(${adatatype} eq "base:integer") {$item="adms_integertostring(mynew$ename->_$aname)";}
      elsif(${adatatype} eq "base:real") {$item="adms_doubletostring(mynew$ename->_$aname)";}
      elsif(${adatatype} eq "*") {$item="adms_admsxml_uid(mynew$ename->_$aname)";}
      elsif(${adatatype} eq "enumeration")
      {
        foreach(@{$_->findnodes("evalue")})
        {
          my$enumname=&attname($_);
          my$enuminfo=&attinfo($_);
          my$info="";
          while($enuminfo=~s/([^\$]*)\$(\d+)//)
          {
            my$i=$2;
            my$v=&attname($uidattributes[$i-1]);
            $info.="    adms_k2strconcat(&myuid,\"$1\");\n";
            $info.="    adms_strconcat(&myuid,adms_admsxml_uid(mynew$ename->_$v));\n";
          }
          $info.="    adms_k2strconcat(&myuid,\"$enuminfo\");\n";
          $enum.="  if(mynew$ename->_$aname->_value==admse_$enumname)\n";
          $enum.="  {\n";
          $enum.="$info";
          $enum.="  }\n";
        }
      }
      elsif((${attributedatatypefamily} eq "reference")&&(${adatatype} eq "adms")) {$item="adms_admsxml_uid(mynew$ename->_$aname)";}
      elsif(${attributedatatypefamily} eq "reference") {$item="adms_${adatatype}_uid(mynew$ename->_$aname)";}
      else{die "attdatatypefamily=${attributedatatypefamily} adatatype=${adatatype}";}
      if(!$enum)
      {
        $format =~ s/\$$count/"\@$item\@"/g;
      }
    }
    $count++;
  }
  my@strsplit;
  if($format)
  {
    $format="\"$format\"\@";
    $format=~s/""\@//g;
    @strsplit=split"\@",$format;
  }
  $code.="char*adms_${ename}_uid (p_$ename mynew$ename)\n";
  $code.="{\n";
  if(scalar(@strsplit))
  {
    $code.="  char*myuid=NULL;\n";
  }
  else
  {
    $code.="  char*myuid=adms_kclone(\"\");\n";
  }
  $count=0;
  foreach(@strsplit)
  {
     if(/^adms_/)
     {
       $code.="  char*myuid$count=$_;\n";
       $count++;
     }
  }
  $count=0;
  if($enum)
  {
    $code.=$enum;
  }
  else
  {
    foreach(@strsplit)
    {
       if(/^adms_/)
       {
         $code.="  adms_strconcat(&myuid,myuid$count);\n";
         $count++;
       }
       else
       {
         $code.="  adms_k2strconcat(&myuid,$_);\n";
       }
    }
  }
  $code.="  return myuid;\n";
  $code.="}\n";
  return $code;
}
sub parent_new
{
  my$leaf=shift;
  my$leafpath=shift;
  my$element=shift;
  my$parent=shift;
  my$code;
  my$parentName=$parent->getAttribute("name");
  my$superElement=$myElementHash->{$parentName};
  my$supersuperElement=&eltparent($superElement);
  if($supersuperElement)
  {
    my$newcode=&parent_new(${leaf},"${leafpath}_$parentName.",$superElement,$supersuperElement);
    $code.=$newcode;
  }
  my@superElementAttribute=&eltattribute($superElement);
  foreach(@superElementAttribute)
  {
    my$aname=&attname($_);
    my$adatatype=&attdatatypename($_);
    my$attributedatatypefamily=&attdatatypefamily($_);
    my$adefault=&attdefault($_);
    if(${adefault})
    {
      my $defval;
      if(${adefault}=~/^@/)
      {
        ${adefault}=~s/@//;
        $defval="  ${leafpath}_$parentName._${aname}=admse_".${leaf}->getAttribute(${adefault}).";\n";
      }
      else
      {
        if(${attributedatatypefamily} eq "basictype")
        {
          if(${adatatype} eq "base:integer")
          {
            $defval="  ${leafpath}_$parentName._${aname}=${adefault};\n";
          }
          elsif(${adatatype} eq "base:enumeration")
          {
            $defval="  ${leafpath}_$parentName._${aname}=admse_${adefault};\n";
          }
        }
        else
        {
          if(${adatatype} eq "real")
          {
            $defval.="  (${leafpath}_$parentName)._${aname}=adms_real_new(${adefault});\n";
          }
          elsif(${adatatype} eq "enumeration")
          {
            $defval.="  (${leafpath}_$parentName)._${aname}=";
            $defval.="adms_enumeration_new(admse_${adefault});\n";
          }
        }
        $defval="${leafpath}_$parentName._${aname}=\"${adefault}\"" if((${adatatype} eq "quark")&&((${adefault} ne "myfile")));
        $defval="${leafpath}_$parentName._${aname}=${adefault}" if((${adatatype} eq "quark")&&((${adefault} eq "myfile")));
      }
      $code.=$defval;
    }
  }
  return $code;
}

sub miscellaneous_code
{
  my$codeH;
  my$codeC;
  $codeC.="\n/*-- Miscellaneous routines --*/\n";
  $codeC.=q@
void breakpoint(void) {}
char*adms_integertostring(int value)
{
   char*string=malloc(sizeof(char)*50);
   sprintf(string,"%i",value);
   return string;
}
char*adms_doubletostring(double value)
{
   if(isnan(value))
   {
     char*string=malloc(sizeof(char)*4);
     sprintf(string,"nan");
     return string;
   }
   else
   {
     char*string=malloc(sizeof(char)*50);
     sprintf(string,"%e",value);
     return string;
   }
}
int adms_file_isregular(const char* myfilename)
{
  return((access(myfilename,ADMS_F_OK)==0));
}
int adms_file_isdirectory(const char* myfilename)
{
  struct stat s;
  return((stat(myfilename,&s)==0)&&(s.st_mode&ADMS_S_IFDIR));
}
int adms_setenv(const char* myname,const char* myvalue)
{
  int myres;
#if defined(HAVE_PUTENV)
  char*myassign=NULL;
  adms_k2strconcat(&myassign,myname);
  adms_k2strconcat(&myassign,"=");
  adms_k2strconcat(&myassign,myvalue);
  myres=putenv(myassign);
  free(myassign);
#else
  myres=setenv(myname,myvalue,1);
#endif
  return (myres==0);
}

/*i*/p_slist adms_slist_last (p_slist l)
{
  if(l)
  {
    while(l->next)
      l=l->next;
  }
  return l;
}
/*i*/p_slist adms_slist_new (p_adms d)
{
  p_slist newl=NULL;
  adms_slist_push(&newl,d);
  return newl;
}
/*i*/p_slist adms_slist_copy (p_slist l)
{
  p_slist copiedl=NULL;
  while(l)
  {
    adms_slist_push(&copiedl,l->data);
    l=l->next;
  }
  return adms_slist_reverse(copiedl);
}
/*i*/void adms_slist_push(p_slist* l,p_adms data)
{
  p_slist n=malloc(sizeof(t_slist));
  n->next=*l;
  n->data=data;
  *l=n;
}
/*i*/p_adms adms_slist_pull(p_slist* l)
{
  if(*l)
  {
    p_slist n=*l;
    p_adms data=n->data;
    *l=(*l)->next;
    free(n);
    return data;
  }
  return NULL;
}
/*i*/void adms_slist_concat (p_slist* l1,p_slist l2)
{
  if(l2)
  {
    if(*l1)
      adms_slist_last(*l1)->next=l2;
    else
      *l1=l2;
  }
}
/*i*/unsigned int adms_slist_length (p_slist l)
{
  unsigned int length=0;
  while(l)
  {
    length++;
    l=l->next;
  }
  return length;
}
/*i*/p_slist adms_slist_nth (p_slist l,unsigned int  n)
{
  while (n-->0 && l)
    l=l->next;
  return l;
}
/*i*/p_adms adms_slist_nth_data (p_slist l,unsigned int n)
{
  while (n-->0 && l)
    l=l->next;
  return l ? l->data : ((p_adms)0);
}
/*i*/p_slist adms_slist_find (p_slist l,p_kadms data)
{
  while(l)
  {
    if(l->data==data)
      break;
    l=l->next;
  }
  return l;
}
/*i*/int adms_slist_index (p_slist l, p_kadms data)
{
  int i=0;
  while(l)
  {
    if(l->data==data)
      return i;
    i++;
    l=l->next;
  }
  return -1;
}
/*i*/p_slist adms_slist_reverse (p_slist l)
{
  p_slist p=NULL;
  while(l)
  {
    p_slist n=l->next;
    l->next=p;
    p=l;
    l=n;
  }
  return p;
}
/*i*/void adms_slist_inreverse (p_slist* l)
{
  *l=adms_slist_reverse(*l);
}
/*i*/void adms_slist_free (p_slist l)
{
  while(l)
  {
    p_slist freed=l;
    l=l->next;
    free(freed);
  }
}
@;
  $codeC.="int globalnbadmstnew=0, globalnbadmstdestroy=0;\n";
  $codeC.="inline int adms_global_nbadmstnew(void) {return globalnbadmstnew;};\n";
  $codeC.="inline int adms_global_nbadmstdestroy(void) {return globalnbadmstdestroy;};\n";
  $codeC.="inline void adms_global_increment_nbadmstnew(void) {globalnbadmstnew++;};\n";
  $codeC.="inline void adms_global_increment_nbadmstdestroy(void) {globalnbadmstdestroy++;};\n";
  $codeH.="win32_interface inline int adms_global_nbadmstnew(void);\n";
  $codeH.="win32_interface inline int adms_global_nbadmstdestroy(void);\n";
  $codeH.="win32_interface inline void adms_global_increment_nbadmstnew(void);\n";
  $codeH.="win32_interface inline void adms_global_increment_nbadmstdestroy(void);\n";

  $codeC.="p_admsmain globaladmsmain;\n";
  $codeC.="p_admsmain root(void) {return globaladmsmain;}\n";

  $codeC.="void rootnew(p_admsmain myglobaladmsmain) {globaladmsmain=myglobaladmsmain;}\n";
  $codeC.="\n/*-- Messaging --*/\n";
  $codeC.="\n";
sub message
{
  my($name,$ofh,$isfatal)=(shift,shift,shift);
  my$codeC;
  my$prefix=sprintf("\"[%-7s]",$name);
  $prefix=~s/ /\./g;
  $prefix.=" \"";
  $codeC.="_t_message (adms_message_${name}_impl)
{
  va_list ap;
  int insideformat=0;
  int i;
  char* s;
  double d;
  void* p;
  fputs($prefix,$ofh);
  va_start(ap, format);
  for(;*format;format++)
  {
    if(insideformat)
    {
      insideformat=0;
      switch(*format) 
      {
        case 's':
          s=va_arg (ap,char*);
          fputs(s,$ofh);
          break;
        case 'e':
          d=va_arg (ap,double);
          fprintf($ofh,\"%e\",d);
          break;
        case 'g':
          d=va_arg (ap,double);
          fprintf($ofh,\"%g\",d);
          break;
        case 'f':
          d=va_arg (ap,double);
          fprintf($ofh,\"%f\",d);
          break;
        case 'i':
          i=va_arg (ap,int);
          fprintf($ofh,\"%i\",i);
          break;
        case 'p':
          p=va_arg (ap,void*);
          fprintf($ofh,\"%p\",p);
          break;
        default:
          fputc(*format,$ofh);
      }
    }
    else
    {
      switch(*format) 
      {
        case '%':
          insideformat=1;
          break;
        default:
          fputc(*format,$ofh);
      }
    }
  }
  va_end (ap);
  fflush($ofh);
";
  if($isfatal)
  {
    $codeC.="  breakpoint();\n";
    $codeC.="  exit(1);\n";
  }
  $codeC.="}\n";
  $codeC.="_t_message (adms_message_${name}_continue_impl)
{
  va_list ap;
  int insideformat=0;
  int i;
  char* s;
  double d;
  void* p;
";
  if($isfatal)
  {
    $codeC.="  fputs($prefix,$ofh);\n";
  }
  $codeC.="  va_start(ap, format);
  for(;*format;format++)
  {
    if(insideformat)
    {
      insideformat=0;
      switch(*format) 
      {
        case 's':
          s=va_arg (ap,char*);
          fputs(s,$ofh);
          break;
        case 'e':
          d=va_arg (ap,double);
          fprintf($ofh,\"%e\",d);
          break;
        case 'g':
          d=va_arg (ap,double);
          fprintf($ofh,\"%g\",d);
          break;
        case 'f':
          d=va_arg (ap,double);
          fprintf($ofh,\"%f\",d);
          break;
        case 'i':
          i=va_arg (ap,int);
          fprintf($ofh,\"%i\",i);
          break;
        case 'p':
          p=va_arg (ap,void*);
          fprintf($ofh,\"%p\",p);
          break;
        default:
          fputc(*format,$ofh);
      }
    }
    else
    {
      switch(*format) 
      {
        case '%':
          insideformat=1;
          break;
        default:
          fputc(*format,$ofh);
      }
    }
  }
  va_end (ap);
  fflush($ofh);
}\n";
  return $codeC;
}  
  $codeH.="\n/*-- Miscellaneous routines --*/\n";
  $codeH.=q@
typedef struct s_slist t_slist;
typedef t_slist* p_slist;
#define adms_slist(member) ((p_slist)(member))
/*typedef struct s_str t_str;*/
typedef char* p_str;
typedef const p_str p_kstr;
#define adms_str(member) ((p_str)(member))
/*d*/win32_interface void breakpoint(void);
/*d*/win32_interface char*adms_integertostring(int value);
/*d*/win32_interface char*adms_doubletostring(double value);
/*d*/win32_interface int adms_setenv(const char* myname,const char* myvalue);
/*d*/win32_interface int adms_file_isregular(const char* myfilename);
/*d*/win32_interface int adms_file_isdirectory(const char* myfilename);
/*d*/win32_interface p_slist adms_slist_new (p_adms d);
/*d*/win32_interface p_slist adms_slist_copy (p_slist l);
/*d*/win32_interface void adms_slist_concat (p_slist* l1,p_slist l2);
/*d*/win32_interface p_slist adms_slist_find (p_slist l,p_kadms data);
/*d*/win32_interface void adms_slist_free (p_slist l);
/*d*/win32_interface int adms_slist_index (p_slist l, p_kadms data);
/*d*/win32_interface p_slist adms_slist_last (p_slist l);
/*d*/win32_interface unsigned int adms_slist_length (p_slist l);
/*d*/win32_interface p_slist adms_slist_nth (p_slist l,unsigned int  n);
/*d*/win32_interface p_adms adms_slist_nth_data (p_slist l,unsigned int n);
/*d*/win32_interface void adms_slist_print(const char* message,p_slist l);
/*d*/win32_interface p_adms adms_slist_pull(p_slist* l);
/*d*/win32_interface void adms_slist_push(p_slist* l,p_adms data);
/*d*/win32_interface p_slist adms_slist_reverse (p_slist l);
/*d*/win32_interface void adms_slist_inreverse (p_slist* l);
struct s_slist
{
  p_adms data;
  p_slist next;
};
@;

  $codeH.="\n";
  $codeH.="win32_interface p_admsmain root(void);\n";
  $codeH.="win32_interface void rootnew(p_admsmain myglobaladmsmain);\n";
  $codeH.="\n";
  $codeH.="typedef p_adms (t_new) (p_adms arguments[]);\n";
  $codeH.="typedef p_adms (*p_new) (p_adms arguments[]);\n";
  $codeH.="#define _t_new(function) p_adms (function) (p_adms arguments[])\n";
  $codeH.="\n";
  $codeH.="typedef const char* (t_returnstring) (void);\n";
  $codeH.="typedef const char* (*p_returnstring) (void);\n";
  $codeH.="#define _t_returnstring(function) const char* (function) (void)\n";
  $codeH.="\n";
  $codeH.="typedef int (t_cmp) (p_adms myadms, p_adms myadmsref);\n";
  $codeH.="typedef int (*p_cmp) (p_adms myadms, p_adms myadmsref);\n";
  $codeH.="#define _t_cmp(function) int (function) (p_adms myadms, p_adms myadmsref)\n";
  $codeH.="\n/*-- Messaging --*/\n";
  $codeH.="\n";
  $codeH.="typedef void (t_message) (const char*format, ...);\n";
  $codeH.="typedef void (*p_message) (const char*format, ...);\n";
  $codeH.="#define _t_message(function)  void (function) (const char*format, ...)\n";
  my@message=(
    ["info","stdout",0],
    ["usage","stdout",0],
    ["verbose","stdout",0],
    ["debug","stdout",0],
    ["dbg_vla","stdout",0],
    ["dbg_xml","stdout",0],
    ["dbg_mem","stdout",0],
    ["hint","stderr",0],
    ["warning","stderr",0],
    ["obsolete","stderr",0],
    ["error","stderr",0],
    ["fatal","stderr",1]
  );
  $codeH.=q[
win32_interface inline void adms_toupper(char*m);
win32_interface inline void adms_tolower(char*m);
win32_interface inline char*adms_consttoupper(const char* m);
win32_interface inline char*adms_consttolower(const char* m);
win32_interface inline char*adms_kclone(const char* m);
win32_interface inline char*adms_knclone(const char* m,const int l);
win32_interface inline char*adms_m2nclone(const char* m,const char* n);
win32_interface inline void adms_strconcat(char **s1,char *s2);
win32_interface inline void adms_k2strconcat(char **s1,const char* s2);
win32_interface FILE *adms_file_open_read (const char* fileName);

];
  $codeC.=q[
inline char*adms_consttoupper(const char* m)
{
  int l=strlen(m);
  char*mycpy=(char*)malloc((l+1)*sizeof(char));
  mycpy[l--]='\0';
  for(;l>=0;l--)
    mycpy[l]=toupper(m[l]);
  return mycpy;
}
inline char*adms_consttolower(const char* m)
{
  int l=strlen(m);
  char*mycpy=(char*)malloc((l+1)*sizeof(char));
  mycpy[l--]='\0';
  for(;l>=0;l--)
    mycpy[l]=tolower(m[l]);
  return mycpy;
}
inline void adms_toupper(char*m)
{
  for(;*m;m++)
    *m=toupper(*m);
}
inline void adms_tolower(char*m)
{
  for(;*m;m++)
    *m=tolower(*m);
}
inline char*adms_kclone(const char* m)
{
  if(m)
  {
    int l=strlen(m);
    char*mycpy=(char*)malloc((l+1)*sizeof(char));
    memcpy(mycpy,m,l);
    mycpy[l]='\0';
    return mycpy;
  }
  else
    return NULL;
}
inline char*adms_knclone(const char* m,const int l)
{
  char*mycpy=(char*)malloc((l+1)*sizeof(char));
  memcpy(mycpy,m,l);
  mycpy[l]='\0';
  return mycpy;
}
/*clone array (m,n(: "abcdefghijkl\0" returns as "cdef\0"*/
/*                      m   n                            */
/* same as adms_knclone(m,n-m)                       */
inline char*adms_m2nclone(const char* m,const char* n)
{
  int l=(char*)n-(char*)m;
  char*mycpy=(char*)malloc((l+1)*sizeof(char));
  memcpy(mycpy,m,l);
  mycpy[l]='\0';
  return mycpy;
}
/*s1: realloced s2: constant ret: s1=s1s2*/
inline void adms_k2strconcat(char **s1,const char* s2)
{
  if(*s1)
  {
    int l1=strlen(*s1);
    int l2=strlen(s2);
    *s1=(char*)realloc(*s1,(l1+l2+1)*sizeof(char));
    memcpy(*s1+l1,s2,l2+1);
  }
  else
    *s1=strdup(s2);
}
/*s1: realloced s2: freed ret: s1=s1s2*/
inline void adms_strconcat(char **s1,char *s2)
{
  adms_k2strconcat(s1,s2);
  free(s2);
}
FILE* adms_file_open_read (const char* myfilename)
{
  FILE* fh=fopen(myfilename,"r");
  if(!fh)
    adms_message_fatal(("%s: failed to open file [read mode]\n",myfilename))
  return fh;
}
];
  $codeC.="\n";
  foreach my$mess(@message)
  {
    my$name=$mess->[0];
    my$outputfile=$mess->[1];
    my$isfatal=$mess->[2];
    $codeC.=message($name,$outputfile,$isfatal);
    $codeH.="win32_interface t_message adms_message_${name}_impl;\n";
    $codeH.="win32_interface t_message adms_message_${name}_continue_impl;\n";
    $codeH.="#define adms_message_${name}(VAARGS)\\\n";
    $codeH.="{\\\n";
    $codeH.="  if((root())\\\n";
    $codeH.="    &&(root()->_${name}->_value==admse_yes))\\\n";
    $codeH.="    adms_message_${name}_impl VAARGS;\\\n";
    $codeH.="}\n";
    $codeH.="#define adms_message_${name}_continue(VAARGS)\\\n";
    $codeH.="{\\\n";
    $codeH.="  if((root())\\\n";
    $codeH.="    &&(root()->_${name}->_value==admse_yes))\\\n";
    $codeH.="    adms_message_${name}_continue_impl VAARGS;\\\n";
    $codeH.="}\n";
  }
  $codeH.="\n";
  return ($codeH,$codeC);
}
sub build_admsxml_accessors
{
  my$codeH;
  my$codeC;
  $codeC.="/*-- accessors: grap data from adms.xml based on adms element name and attribute name --*/\n";
  $codeC.="#include <admsElement/admsC.h>\n";
  $codeC.="\n";
  {
    $datatypename="  int dataTypeName=myitem->_datatypename;\n";
    $codeC.="char* adms_admsxml_uid (p_kadms myitem)\n";
    $codeC.="{\n";
    $codeC.=$datatypename;
    foreach(@$myElementArray)
    {
      my$ename=$_->getAttribute("name");
      next if $ename eq "adms";
      $codeC.="  if(dataTypeName==admse_$ename)\n";
      $codeC.="    return adms_${ename}_uid(adms_$ename(myitem));\n";
    }
    {
      $codeC.="  {\n";
      $codeC.="    adms_message_fatal_continue((\"%s:%i: internal error - should not be reached\\n\",__FILE__,__LINE__))\n";
      $codeC.="    adms_message_fatal((\"%s:%i: please report to r29173\@users.sourceforge.net\\n\",__FILE__,__LINE__))\n";
      $codeC.="  }\n";
    }
    {$codeC.="  return NULL;\n";}
    $codeC.="}\n";
  }
  return ($codeH,$codeC);
}
